#!/bin/bash

## [Goal] BfR-ABC-Harmonization - deploy_all_pipelines.sh
## [Author] Holger Brendebach

## Shell Behaviour --------------------------------------------------
set -Euo pipefail # to exit immediately, add e

## (1) Argument Logic

usage() {
  cat <<-USE

${bold}Basic usage:${normal}
bash $script_name [Options]

${bold}Description:${normal}
Deploy all pipelines in the scope of the BfR-ABC-Harmonization repository

${bold}Optional Named Arguments:${normal}
-d, --deployment                Deploy BfR-ABC pipelines to <deploy_path>
-a, --analysis                  Analyse BfR-ABC pipelines test datasets
-p, --deploy_path               Override deploy base path (default: <git_parent_dir> from paths.cfg)
--auto                          Do not ask for interactive confirmation, e.g. for automated deployments
--imeanit                       Do not perform a dryrun

${bold}Example:${normal}
bash $script_real --deploy_paths $(dirname "$repo_path") --auto --imeanit

USE
  exit 0
}

parse_args() {
  local args=("$@")

  ## default values of variables
  repo_path=$(dirname "$script_path")
  args_git_parent_dir=""
  args_dryrun="--dryrun"
  args_interactive=""

  ## default values of switches
  args_deployment=false
  args_analysis=false
  dryrun=true
  interactive=true

  while :; do
    case "${1-}" in
    -h | --help) usage ;;
    --dev) dev_notes "$script_real" ;; # from helper_functions.sh
    -v | --verbose)
      set -x
      shift
      ;;
    -d | --deployment)
      args_deployment=true
      shift
      ;;
    -a | --analysis)
      args_analysis=true
      shift
      ;;
    -p | --deploy_paths)
      args_git_parent_dir=${2:-}
      shift 2
      ;;
    -imeanit | --imeanit)
      args_dryrun=""
      dryrun=false
      shift
      ;;
    --auto) # -a is already taken
      args_interactive="--auto"
      interactive=false
      shift
      ;;
    -?*) die "Unknown option: $1. For the manual, type ${blue}bash $script_real --help${normal}" ;;
    *) break ;;
    esac
  done

  [[ $dryrun == true ]] && echo -e "\n${bold}${green}-->  This is a dryrun with no log  <--${normal}\n"
  [[ $dryrun == false ]] && echo -e "\n${bold}${magenta}-->  This is NOT a dryrun!  <--${normal}\n"

  return 0
}

## Source Helper Functions --------------------------------------------------
script_real=$(realpath "${BASH_SOURCE[0]}")
script_path=$(dirname "$script_real")
script_name=$(basename "$script_real")
[[ ! -f "$script_path/helper_functions.sh" ]] && echo "[ERROR] File not found: $script_path/helper_functions.sh" && kill -s TERM $$
# shellcheck source=./helper_functions.sh
source "$script_path/helper_functions.sh"

## Parse Arguments --------------------------------------------------
args="$*"
parse_args "$@"

## (2) Script Logic

## Notification on unexpected behaviour
[[ $(declare -F error_handler) == error_handler ]] && msg "Email notification enabled" && trap error_handler SIGINT SIGTERM ERR

## Define Paths --------------------------------------------------

define_paths() {
  ## [Rationale] define_paths() use real-paths for proper awk splitting

  ## resources
  config_file=$repo_path/static/paths.cfg
  pipeline_table=$repo_path/static/pipelines.tsv

  ## checks
  [[ ! -f $config_file ]] && die "File not found: $config_file"
  [[ ! -f $pipeline_table ]] && die "File not found: $pipeline_table"

  ## source a config file
  source "$config_file"
  [[ -n $args_git_parent_dir ]] && git_parent_dir=$(realpath "$args_git_parent_dir")
  [[ ! -d $git_parent_dir ]] && logexec mkdir -p "$git_parent_dir"

  ## logging
  logfile_name=${script_name/.sh/.log}
  [[ $dryrun == true ]] && logfile=/dev/null         # ADD # NOTE: defaults to /tmp except for dry-runs (this line), define a file and test for dryrun=false for instance logging
  [[ $dryrun == false ]] && logfile_global=/dev/null # ADD # NOTE: defaults /dev/null, define a file for global logging

  return 0
}

## Info --------------------------------------------------

show_info() {
  logheader "Parameters:"
  echo "
git_parent_dir:|${git_parent_dir:-not sourced}
Dryrun:|$dryrun
Logfile:|$logfile
" | column -t -s "|" | logpipe

  logheader "Harmonized Pipelines:"
  mlr --t2p cat "$pipeline_table"
}

## Functions ------------------------------------------------

pipeline_table_column2array() {
  local column_name column_index column_content
  column_name=$1

  mapfile -t column_index < <(mlr --tsv --headerless-csv-output cut -f "repository_name" "$pipeline_table") || die "Could not read column: repository_name"
  mapfile -t column_content < <(mlr --tsv --headerless-csv-output cut -f "$column_name" "$pipeline_table") || die "Could not read column: $column_name"

  declare -g -A column_array
  for i in "${!column_index[@]}"; do column_array["${column_index[$i]}"]="${column_content[$i]}"; done
}

## Modules --------------------------------------------------

deploy_single_pipeline() {
  ## [Rationale] deploy_single_pipeline() sequential switch logic: skip is by default on true and when specific condition is met, the switch is set to the value of dryrun, hence dryrun overrides

  local pipeline branch pipeline_dir skip_git_pull skip_setup_env skip_setup_data skip_test_analysis
  skip_git_pull=true
  skip_setup_env=true
  skip_setup_data=true

  pipeline=$1
  branch=${2:-main}
  logheader "Deploying pipeline $pipeline:"

  ## check if directory exists
  pipeline_dir="$git_parent_dir/$pipeline"

  echo "
pipeline_dir:|$pipeline_dir
branch:|$branch
" | column -t -s "|" | logpipe

  ## clone repository
  [[ ! -d $pipeline_dir ]] && skip_git_pull=false || logwarn "Directory already exists: $pipeline_dir"
  [[ $skip_git_pull == false ]] && (
    logexec git -C "$git_parent_dir" clone --branch "$branch" "${repository_origin_urls[$pipeline]}"
    [[ ! -d $pipeline_dir ]] && die "Directory was not created: $pipeline_dir"
    true
  )

  ## setup mamba environment
  [[ ! "${conda_envs[*]}" =~ ${pipeline,,} && -f "$pipeline_dir/${pipeline_setup_sh[$pipeline]}" ]] && skip_setup_env=false || logwarn "Conda environment already exists or setup script not found"
  [[ $skip_setup_env == false ]] && logexec bash "$pipeline_dir"/"${pipeline_setup_sh[$pipeline]}" $args_interactive "${setup_args_env[$pipeline]}"

  ## setup data dependencies
  [[ -f "$pipeline_dir/${pipeline_setup_sh[$pipeline]}" && $(find $pipeline_dir/test_data -type f \( -name "*.sha256" -o -name "*.md5" \) | wc -l) -lt 1 ]] && skip_setup_data=false || logwarn "Data already retrieved or setup script not found"
  [[ $skip_setup_data == false ]] && logexec bash "$pipeline_dir"/"${pipeline_setup_sh[$pipeline]}" $args_interactive "${setup_args_data[$pipeline]}"

  logsuccess "Pipeline $pipeline deployed successfully"
}

deploy_bfr_abc() {
  ## [Rationale] deploy_bfr_abc() This is the Rationale for deploy_bfr_abc() and mode "--deployment"

  local pipeline branch="feature/harmonization"

  # extract from table
  pipeline_table_column2array "repository_origin_url"
  declare -g -A repository_origin_urls
  for key in "${!column_array[@]}"; do repository_origin_urls["$key"]="${column_array[$key]}"; done

  pipeline_table_column2array "pipeline_setup_sh"
  declare -g -A pipeline_setup_sh
  for key in "${!column_array[@]}"; do pipeline_setup_sh["$key"]="${column_array[$key]}"; done

  pipeline_table_column2array "setup_args_env"
  declare -g -A setup_args_env
  for key in "${!column_array[@]}"; do setup_args_env["$key"]="${column_array[$key]}"; done

  pipeline_table_column2array "setup_args_data"
  declare -g -A setup_args_data
  for key in "${!column_array[@]}"; do setup_args_data["$key"]="${column_array[$key]}"; done

  pipeline_table_column2array "pipeline_wrapper_sh"
  declare -g -A pipeline_wrapper_sh
  for key in "${!column_array[@]}"; do pipeline_wrapper_sh["$key"]="${column_array[$key]}"; done

  logheader "Git Remote Origin URL:"
  column -t -s "|" < <(for key in "${!repository_origin_urls[@]}"; do echo "$key|${repository_origin_urls[$key]}"; done)

  # get installed conda envs
  mapfile -t conda_envs < <(conda env list --json | jq -r '.envs | join("\n")')

  # deploy pipelines from branch
  for pipeline in "${!repository_origin_urls[@]}"; do
    [[ $interactive == true ]] && (
      echo
      interactive_query "Analyse pipeline ${magenta}$pipeline${normal}?" || continue # TODO: interactive_query exits upon "no"; This only exits the subshell caused by the (..) group.
    )
    deploy_single_pipeline "$pipeline" "$branch"
  done

  echo
}

analyse_single_pipeline() {
  ## [Rationale] analyse_single_pipeline() This is the Rationale for analyse_single_pipeline() which is looped in analyze_bfr_abc()

  local pipeline pipeline_dir skip_test_analysis
  skip_test_analysis=true

  pipeline=$1
  logheader "Test data analysis of pipeline $pipeline:"

  ## check if directory exists
  pipeline_dir="$git_parent_dir/$pipeline"

  echo "
pipeline_dir:|$pipeline_dir
" | column -t -s "|" | logpipe

  ## analyse test data
  [[ $(find "$pipeline_dir/test_data" -type f -name "pipeline_status*" | wc -l) -lt 1 ]] && skip_test_analysis=false || logwarn "pipeline_status file found. Skipping pipeline analysis: $pipeline"
  [[ $skip_test_analysis == false ]] && logexec bash "$pipeline_dir"/"${pipeline_wrapper_sh[$pipeline]}" --auto $args_dryrun # --norun

  logsuccess "Pipeline $pipeline test data analysed successfully"
}

analyze_bfr_abc() {
  ## [Rationale] analyze_bfr_abc() This is the Rationale for deploy_bfr_abc() and mode "--analysis"

  local pipeline

  pipeline_table_column2array "pipeline_wrapper_sh"
  declare -g -A pipeline_wrapper_sh
  for key in "${!column_array[@]}"; do pipeline_wrapper_sh["$key"]="${column_array[$key]}"; done

  # deploy pipelines from branch
  for pipeline in "${!pipeline_wrapper_sh[@]}"; do
    [[ $interactive == true ]] && (
      echo
      interactive_query "Analyze data with pipeline ${magenta}$pipeline${normal}?" || continue # TODO: interactive_query exits upon "no"; This only exits the subshell caused by the (..) group.
    )
    analyse_single_pipeline "$pipeline"
  done

  echo
}

## (3) Main

define_paths
logglobal "[START] $script_real"
logglobal "Arguments: ${*:-"No arguments provided"}"
show_info

## Workflow
[[ $args_deployment == true ]] && deploy_bfr_abc
[[ $args_analysis == true ]] && analyze_bfr_abc

logglobal "All steps were logged to: $logfile"
logglobal "[STOP] $script_real"
echo "Thank you for using the BfR-ABC"
