#!/bin/bash

## [Goal] BfR-ABC-Harmonization - create_symlinks.sh
## [Author] Holger Brendebach

## Shell Behaviour --------------------------------------------------
set -Euo pipefail # to exit immediately, add e

## (1) Argument Logic

usage() {
  cat <<-USE

${bold}Basic usage:${normal}
bash $script_name [Options] # ADD

${bold}Description:${normal}
Create or refresh the symlinks in the BfR-ABC Harmonization project.
The default mode is 'query' to create symlinks.

${bold}Optional Named Arguments:${normal}
-q, --query_repositories        Query repository information (mode=query)
-c, --create_symlinks           Create symlinks for harmonization (mode=create)
-n, --dryrun                    Perform a dryrun to review the commands (default=false)

${bold}Example:${normal}
bash $script_real

USE
  exit 0
}

parse_args() {
  local args=("$@")

  ## default values of variables
  repo_path=$(dirname "$script_path")
  mode="query"

  ## default values of switches
  dryrun=false

  while :; do
    case "${1-}" in
    -h | --help) usage ;;
    --dev) dev_notes "$script_real" ;; # from helper_functions.sh
    -v | --verbose)
      set -x
      shift
      ;;
    -q | --query_repositories)
      mode="query"
      shift
      ;;
    -c | --create_symlinks)
      mode="create"
      shift
      ;;
    -n | --dryrun)
      dryrun=true
      shift
      ;;
    -?*) die "Unknown option: $1. For the manual, type ${blue}bash $script_real --help${normal}" ;;
    *) break ;;
    esac
  done

  [[ $dryrun == true ]] && echo -e "\n${bold}${green}-->  This is a dryrun with no log  <--${normal}\n"
  [[ $dryrun == false ]] && echo -e "\n${bold}${magenta}-->  This is NOT a dryrun!  <--${normal}\n"

  return 0
}

## Source Helper Functions --------------------------------------------------
script_real=$(realpath "${BASH_SOURCE[0]}")
script_path=$(dirname "$script_real")
script_name=$(basename "$script_real")
[[ ! -f "$script_path/helper_functions.sh" ]] && echo "[ERROR] File not found: $script_path/helper_functions.sh" && kill -s TERM $$
# shellcheck source=./helper_functions.sh
source "$script_path/helper_functions.sh"

## Parse Arguments --------------------------------------------------
args="$*"
parse_args "$@"

## (2) Script Logic

## Notification on unexpected behaviour
[[ $(declare -F error_handler) == error_handler ]] && msg "Email notification enabled" && trap error_handler SIGINT SIGTERM ERR

## Define Paths --------------------------------------------------

define_paths() {
  ## [Rationale] define_paths() use real-paths for proper awk splitting

  ## resources
  config_file=$repo_path/static/paths.cfg
  pipeline_table=$repo_path/static/pipelines.tsv

  ## checks
  [[ ! -f $config_file ]] && die "File not found: $config_file"
  [[ ! -f $pipeline_table ]] && die "File not found: $pipeline_table"

  ## source a config file
  source $config_file

  ## logging
  logfile_name=${script_name/.sh/.log}
  [[ $dryrun == true ]] && logfile=/dev/null         # ADD # NOTE: defaults to /tmp except for dry-runs (this line), define a file and test for dryrun=false for instance logging
  [[ $dryrun == false ]] && logfile_global=/dev/null # ADD # NOTE: defaults /dev/null, define a file for global logging

  return 0
}

## Info --------------------------------------------------

show_info() {
  logheader "Parameters:"
  echo "
mode:|$mode
git_parent_dir:|${git_parent_dir:-not sourced}
dryrun:|$dryrun
" | column -t -s "|" | tee >(logglobal 1>/dev/null) && sleep 0.2

  logheader "Harmonized Pipelines:"
  mlr --t2p cat "$pipeline_table" # --allow-ragged-csv-input
}

## Functions ------------------------------------------------

pipeline_table_column2array() {
  local column_name column_index column_content
  column_name=$1

  mapfile -t column_index < <(mlr --tsv --headerless-csv-output cut -f "repository_name" "$pipeline_table") || die "Could not read column: repository_name"
  mapfile -t column_content < <(mlr --tsv --headerless-csv-output cut -f "$column_name" "$pipeline_table") || die "Could not read column: $column_name"

  declare -g -A column_array
  for i in "${!column_index[@]}"; do column_array["${column_index[$i]}"]="${column_content[$i]}"; done
}

create_relative_symlink() {
  local target symlink
  target=$1
  symlink=$2

  # checks
  symlink_dir=$(dirname "$symlink")
  [[ ! -d $symlink_dir ]] && logexec mkdir -p "$symlink_dir"
  cd "$symlink_dir" || die "Could not change to symlink directory: $symlink_dir"
#  [[ -f $symlink ]] && logwarn "The target file already exists: $symlink"
  [[ -f $symlink && ! -L $symlink ]] && logerror "The target is not a symlink: $symlink" && return 0

  target_dir=$(dirname "$target")
  [[ ! -d $target_dir ]] && logerror "The target directory does not exist: $target_dir" && return 0
  [[ ! -f $target ]] && logerror "The target directory does not exist: $target_dir" && return 0

  # create link
  logexec "cd \"$symlink_dir\"; ln -sf \"$target\" \"$(basename "$symlink")\""
}

create_absolute_symlink() {
  local target symlink
  target=$1
  symlink=$2

  # checks
  symlink_dir=$(dirname "$symlink")
  [[ ! -d $symlink_dir ]] && logexec mkdir -p "$symlink_dir"
#  [[ -f $symlink ]] && logwarn "The target file already exists: $symlink"
  [[ -f $symlink && ! -L $symlink ]] && logerror "The target is not a symlink: $symlink" && return 0

  target_dir=$(dirname "$target")
  [[ ! -d $target_dir ]] && logerror "The target directory does not exist: $target_dir" && return 0
  [[ ! -f $target ]] && logerror "The target directory does not exist: $target_dir" && return 0

  # create link
  logexec "cd \"$symlink_dir\"; ln -sf \"$target\" \"$(basename "$symlink")\""
}

create_absolute_symlinks_from_array() {
  # init variables and associative array
  local file_array_name
  file_array_name=$1
  declare -A file_array

  # copy associative array by indirect expansion
  for key in $(eval echo \${!$file_array_name[@]}); do eval "file_array[\$key]=\"\${$file_array_name[\$key]}\""; done

  # create symlinks
  for key in "${!file_array[@]}"; do
    create_absolute_symlink "${file_array[$key]}" "$repo_path/code_harmonization/$file_array_name/${key,,}_$(basename ${file_array[$key]})"
  done
}

show_git_origin_url() {
  # init variables and associative array
  local pipeline_dir repository_origin_url
  declare -g -A repository_origin_urls

  # retrieve information
  for pipeline in "${repository_names[@]}"; do
    pipeline_dir=$git_parent_dir/$pipeline
    [[ ! -d $pipeline_dir ]] && logwarn "Directory not found: $pipeline_dir"

    # get remote origin URL
    repository_origin_url=$(git -C "$pipeline_dir" remote get-url origin || die "git command failed for $pipeline")
    repository_origin_urls["$pipeline"]="$repository_origin_url"
  done

  # print
  column -t -s "|" < <(for key in "${!repository_origin_urls[@]}"; do echo "$key|${repository_origin_urls[$key]}"; done)
}

show_git_version() {
  # init variables and associative array
  local pipeline_dir repository_origin_url
  declare -g -A repository_origin_urls

  # retrieve information
  for pipeline in "${repository_names[@]}"; do
    pipeline_dir=$git_parent_dir/$pipeline
    [[ ! -d $pipeline_dir ]] && logwarn "Directory not found: $pipeline_dir"

    # get remote origin URL
    repository_origin_url=$(git -C "$pipeline_dir" describe --always || die "git command failed for $pipeline")
    repository_origin_urls["$pipeline"]="$repository_origin_url"
  done

  # print
  column -t -s "|" < <(for key in "${!repository_origin_urls[@]}"; do echo "$key|${repository_origin_urls[$key]}"; done)
}

show_readme_head() {
  # init variables and associative array
  local pipeline_dir pipeline_readme_path repository_readme_head
  declare -A repository_readme_heads

  # retrieve information
  for pipeline in "${repository_names[@]}"; do
    pipeline_dir=$git_parent_dir/$pipeline
    [[ ! -d $pipeline_dir ]] && logwarn "Directory not found: $pipeline_dir"

    # get readme markdown file
    pipeline_readme_path=$pipeline_dir/README.md
    repository_readme_head=$(head -n1 $pipeline_readme_path || die "No readme found for $pipeline")
    repository_readme_heads["$pipeline"]="$repository_readme_head"
  done

  # print
  column -t -s "|" < <(for key in "${!repository_readme_heads[@]}"; do echo "$key|${repository_readme_heads[$key]}"; done)
}

find_pipeline_file() {
  # init variables and associative array
  local print_output file_name path_prefix is_suffix skip_find pipeline_dir pipeline_file
  declare -g -A pipeline_files

  # read arguments
  path_prefix=${1:-}
  file_name=${2:-}
  print=${3:-true}

  logheader "Find pipeline file: ${path_prefix:-"."}/$file_name"
  [[ -z $file_name ]] && die "No file name (suffix) provided"
  [[ $file_name == _* || $file_name == \.* ]] && is_suffix=true || is_suffix=false

  # retrieve information
  for pipeline in "${repository_names[@]}"; do
    skip_find=false
    pipeline_dir=$git_parent_dir/$pipeline/$path_prefix
    pipeline_dir=$git_parent_dir/$pipeline/$path_prefix
    [[ ! -d $pipeline_dir ]] && logwarn "Directory not found: $pipeline_dir" && skip_find=true

    # find file
    if [[ $skip_find == false ]]; then
      if [[ $is_suffix == true ]]; then
        pipeline_file=$(find $pipeline_dir -maxdepth 1 -type f -iname "${pipeline,,}$file_name" -ls || echo -n "${yellow}[WARNING]${normal} file not found") # TODO: harmonize to avoid find -iname
      else
        pipeline_file=$(find $pipeline_dir -maxdepth 1 -type f -name "$file_name" -ls || echo -n "${yellow}[WARNING]${normal} file not found")
      fi
    fi
    pipeline_files["$pipeline"]="${pipeline_file:-"${yellow}[WARNING]${normal} file not found"}"
  done

  # print
  [[ $print == true ]] && column -t -s "|" < <(for key in "${!pipeline_files[@]}"; do echo "$key|${pipeline_files[$key]}"; done)

  return 0
}

strip_ls2array() {
  local new_array
  new_array=$1
  declare -g -A "$new_array"

  for key in "${!pipeline_files[@]}"; do
    eval "$new_array[\"$key\"]=\"$(awk '{print $NF}' <<<\"${pipeline_files[$key]})\""
  done
}

## Modules --------------------------------------------------

query_info() {
  ## [Rationale] query_info() This is the Rationale for query_info() and mode "query"

  # get repository names
  pipeline_table_column2array "repository_name"
  repository_names=("${column_array[@]}")

  echo -e "\n${magenta}############################## miscellaneous infos${normal}"

  logheader "Git Remote Origin URL:"
  show_git_origin_url

  logheader "Repository README head:"
  show_readme_head

  echo -e "\n${magenta}############################## root files ${normal}"

  find_pipeline_file "" "VERSION" && strip_ls2array "version_files"
  find_pipeline_file "" "README.md"

  echo -e "\n${magenta}############################## version info ${normal}"

  logheader "Git Version:"
  show_git_version

  logheader "VERSION Version:"
  column -t -s "|" < <(for key in "${!version_files[@]}"; do echo "$key|$(cat ${version_files[$key]})"; done)

  echo -e "\n${magenta}############################## generic scripts${normal}"

  find_pipeline_file "scripts" "helper_functions.sh"
  find_pipeline_file "scripts" "create_sampleSheet.sh"

  echo -e "\n${magenta}############################## specific scripts${normal}"

  find_pipeline_file "" ".py"
  find_pipeline_file "scripts" "_setup.sh"
  find_pipeline_file "scripts" "_wrapper.sh"

  echo
}

create_symlinks() {
  ## [Rationale] create_symlinks() This is the Rationale for create_symlinks() and mode "create"

  logheader "Creating symlinks:"
  create_relative_symlink "../../scripts/helper_functions.sh" "$repo_path/scripts/helper_functions.sh"

  # get repository names
  pipeline_table_column2array "repository_name"
  repository_names=("${column_array[@]}")

  echo -e "\n${magenta}############################## generic scripts${normal}"

  find_pipeline_file "scripts" "helper_functions.sh" false
  strip_ls2array "helper_functions_sh"
  create_absolute_symlinks_from_array "helper_functions_sh"

  find_pipeline_file "scripts" "create_sampleSheet.sh" false
  strip_ls2array "create_sampleSheets_sh"
  create_absolute_symlinks_from_array "create_sampleSheets_sh"

  find_pipeline_file "" ".py" false
  strip_ls2array "pipeline_py"
  create_absolute_symlinks_from_array "pipeline_py"

  find_pipeline_file "scripts" "_setup.sh" false
  strip_ls2array "pipeline_setup_sh"
  create_absolute_symlinks_from_array "pipeline_setup_sh"

  find_pipeline_file "scripts" "_wrapper.sh" false
  strip_ls2array "pipeline_wrapper_sh"
  create_absolute_symlinks_from_array "pipeline_wrapper_sh"

  find_pipeline_file "scripts" "_runs.tsv" false
  strip_ls2array "pipeline_runs_tsv"
  create_absolute_symlinks_from_array "pipeline_runs_tsv"

  find_pipeline_file "profiles/bfr.hpc" "config.yaml" false
  strip_ls2array "smk_profile_yaml"
  create_absolute_symlinks_from_array "smk_profile_yaml"

  find_pipeline_file "" "README.md" false
  strip_ls2array "readme_md"
  create_absolute_symlinks_from_array "readme_md"

  find_pipeline_file "" "VERSION" false
  strip_ls2array "version_file"
  create_absolute_symlinks_from_array "version_file"

  find_pipeline_file "" "LICENSE" false
  strip_ls2array "license_file"
  create_absolute_symlinks_from_array "license_file"

  logheader "Find Main Snakefiles:"
  pipeline_table_column2array "main_snakefile"
  main_snakefile_names=("${column_array[@]}")
  declare -g -A main_snakefiles
  for i in "${!repository_names[@]}"; do main_snakefiles["${repository_names[$i]}"]=$git_parent_dir/${repository_names[$i]}/${main_snakefile_names[$i]}; done
#  column -t -s "|" < <(for key in "${!main_snakefiles[@]}"; do echo "$key|${main_snakefiles[$key]}"; done)
  create_absolute_symlinks_from_array "main_snakefiles"

  echo
}

## (3) Main

define_paths
logglobal "[START] $script_real"
logglobal "Arguments: ${*:-"No arguments provided"}"
show_info

## Workflow
case $mode in
query) query_info ;;
create) create_symlinks ;;
*) die "Unknown mode: $mode" ;;
esac

logglobal "All steps were logged to: $logfile"
logglobal "[STOP] $script_real"
